from typing import Final

from ._types import ScalarType
from .types import (
    Float,
    Float64,
    Int8,
    Int8Enum,
    Int16,
    Int16Enum,
    Int32,
    Int32Enum,
    Int64,
    Int64Enum,
    Uint8,
    Uint8Enum,
    Uint16,
    Uint16Enum,
    Uint32,
    Uint32Enum,
    Uint64,
    Uint64Enum,
)

NoneType = type(None)

ENUM_TYPE_TO_SCALAR_TYPE: Final = {
    Uint8Enum: "Uint8",
    Uint16Enum: "Uint16",
    Uint32Enum: "Uint32",
    Uint64Enum: "Uint64",
    Int8Enum: "Int8",
    Int16Enum: "Int16",
    Int32Enum: "Int32",
    Int64Enum: "Int64",
}

SCALAR_TYPE_TO_PYTHON: Final[dict[ScalarType, type]] = {
    "Uint8": int,
    "Uint16": int,
    "Uint32": int,
    "Uint64": int,
    "Int8": int,
    "Int16": int,
    "Int32": int,
    "Int64": int,
    "Bool": bool,
    "Float32": float,
    "Float64": float,
}

SCALAR_TYPE_TO_DEFAULT: Final[dict[ScalarType, str]] = {
    "float": "0.0",
    "double": "0.0",
    "Float32": "0.0",
    "Float64": "0.0",
    "Uint8": "0",
    "Uint16": "0",
    "Uint32": "0",
    "Uint64": "0",
    "Int8": "0",
    "Int16": "0",
    "Int32": "0",
    "Int64": "0",
    "Bool": "False",
}

HELPER_TYPE_TO_SCALAR_TYPE: Final[dict[type, ScalarType]] = {
    Uint8: "Uint8",
    Uint16: "Uint16",
    Uint32: "Uint32",
    Uint64: "Uint64",
    Int8: "Int8",
    Int16: "Int16",
    Int32: "Int32",
    bool: "Bool",
    Float: "Float32",
    Float64: "Float64",
    Int64: "Int64",
}

SCALAR_TYPE_TO_WIDTH: Final[dict[ScalarType, int]] = {
    "Bool": 1,
    "Uint8": 1,
    "Uint16": 2,
    "Uint32": 4,
    "Uint64": 8,
    "Int8": 1,
    "Int16": 2,
    "Int32": 4,
    "Int64": 8,
    "Float32": 4,
    "Float64": 8,
}



from lark import Lark

parser = Lark(
    r"""
namespace : "namespace" NAMESPACE_NAME ";"

LCASE_LETTER: "a".."z"
UCASE_LETTER: "A".."Z"
DIGIT: "0".."9"
DECIMAL_POSITIVE_INTEGER: "1".."9" DIGIT*

LETTER: UCASE_LETTER | LCASE_LETTER
WORD: LETTER+

COMMENT: "//" /.*/ "\n"
         | "/*" /(.|\n|\r)+/ "*/"

NAMESPACE_NAME: LETTER (LETTER | DIGIT | "_" | ".")*
FILEPATH: (LETTER | DIGIT | "_" | "." | "/")+ ".fbs"

NAME: LETTER (LETTER | DIGIT | "_" )*
NAMESPACED_NAME: LETTER (LETTER | DIGIT | "_" | "." )* (LETTER | DIGIT | "_")+

root_type: "root_type" NAME ";"
include: "include \"" FILEPATH "\"" ";"

enum: "enum" NAME ":" INT_TYPES "{" [ enum_field ("," enum_field)*] ","? "}"
enum_field: NAME enum_field_default?
enum_field_default: "=" NUMBER

table: "table" NAME attributes? "{" table_field* "}"
table_field: NAME ":" (scalar_type | vector_type | NAMESPACED_NAME) table_field_default? attributes? ";"
table_field_default: "=" (NAMESPACED_NAME | NUMBER)
attributes: "(" [WORD ("," WORD)*] ")"

union: "union" NAME "{" [union_member (("," union_member ","?)* | ",")] "}"
union_member: NAMESPACED_NAME ("=" DECIMAL_POSITIVE_INTEGER)?

attribute: "attribute \"" NAME "\"" ";"

BOOL_TYPE: "bool"
INT_TYPES: "u"? "int" ("8"|"16"|"32"|"64")? | "ubyte" | "ushort" | "ulong"
           | "byte" | "short" | "long"
FLOAT_TYPES: "float" | "float32" | "double" | "float64"
STRING_TYPE: "string"
scalar_type: BOOL_TYPE | INT_TYPES | FLOAT_TYPES
vector_type: "[" NAMESPACED_NAME "]"

module: include* namespace? (table | enum | union | attribute)* root_type?

%import common.SIGNED_NUMBER -> NUMBER
%import common.WS
%ignore WS
%ignore COMMENT
""",
    start="module",
)



from enum import Enum as PythonEnum
from keyword import iskeyword
from pathlib import Path
from traceback import print_exception
from typing import Final, Mapping, Sequence, Set, TypeAlias

from attrs import define, evolve, field, frozen
from immutables import Map
from lark import ParseTree, Token, Tree
from lark.visitors import Interpreter

from .._types import FieldName, Optionality
from .parser import parser

Imports: TypeAlias = Mapping[str, Set[str]]
Script: TypeAlias = list[str]
ImportableName: TypeAlias = str
MISSING: Final = "$MISSING"


@frozen
class Attribute:
    name: str
    value: str | None = None
    namespace: str | None = MISSING


BUILTIN_ATTRS = {"required", "deprecated", "key"}
NO_REPR = Attribute("norepr", namespace="flattrs")
FROZEN = Attribute("immutable", namespace="flattrs")


@frozen
class Table:
    @frozen
    class Field:
        name: FieldName
        type: str
        default: str
        is_optional: Optionality
        namespace_prefix: str | None = None
        attrs: Sequence[Attribute] = field(factory=tuple, converter=tuple)

    name: ImportableName
    field_defs: Sequence[Field] = field(converter=tuple)
    imports: Imports = field(converter=Map)
    attrs: Sequence[Attribute] = field(converter=tuple)

    def adjust_enums(self, enums: set[str]) -> "Table":
        """
        Table fields that are enums must not be optional.

        Also, we adjust the default values for enums, if present.

        In addition, we apply the keyword remap to the default, if needed.
        """
        field_defs = [
            f
            if f.type not in enums
            else evolve(
                f,
                is_optional=False,
                default=f"{f.type}.{map_python_keywords(f.default)}"
                if f.default
                else f.default,
            )
            for f in self.field_defs
        ]
        return evolve(self, field_defs=field_defs)

    def adjust_defaults(self) -> "Table":
        """Insert default values as possible, starting from the back.

        This is to preserve backwards comp when optional fields are added.

        This needs to run after `adjust_enums`, since it depends on the
        optionality of fields.
        """
        new_field_defs = []
        can_add_defaults = True
        for field_def in reversed(self.field_defs):
            if can_add_defaults:
                if field_def.is_optional:
                    if not field_def.default:
                        field_def = evolve(field_def, default="None")
                elif not field_def.default:
                    # We need to stop adding defaults.
                    can_add_defaults = False
            new_field_defs.append(field_def)
        return evolve(self, field_defs=list(reversed(new_field_defs)))

    def adjust_attribute_namespaces(
        self, namespace_map: dict[str, set[str]]
    ) -> "Table":
        new_table_attrs = []
        for attr in self.attrs:
            if attr.namespace == MISSING:
                for n, ats in namespace_map.items():
                    if attr.name in ats:
                        attr = evolve(attr, namespace=n)
                        break
                else:
                    raise Exception(f"Couldn't resolve attribute {attr}")
                new_table_attrs.append(attr)
        new_field_defs = []
        for field_def in self.field_defs:
            new_attrs = []
            for attr in field_def.attrs:
                if attr.namespace == MISSING:
                    for n, ats in namespace_map.items():
                        if attr.name in ats:
                            attr = evolve(attr, namespace=n)
                            break
                    else:
                        raise Exception(
                            f"{self.name}: Couldn't resolve attribute {attr}"
                        )
                new_attrs.append(attr)
            new_field_defs.append(evolve(field_def, attrs=new_attrs))
        return evolve(self, attrs=new_table_attrs, field_defs=new_field_defs)

    def adjust_imports(self) -> "Table":
        """This needs to be called after `adjust_attribute_namespace` for correctness."""
        is_frozen = FROZEN in self.attrs
        imports = merge_imports(
            self.imports, {"attrs": {"define"} if not is_frozen else {"frozen"}}
        )
        for f in self.field_defs:
            has_repr = True
            if f.type == "bytes" or NO_REPR in f.attrs:
                has_repr = False
            needs_field = not has_repr
            if needs_field:
                imports = merge_imports(imports, {"attrs": {"field"}})
        return evolve(self, imports=imports)

    def render(self) -> Script:
        # This can be a little tricky, since we need to use `attrs.field`
        # in case we need to supress the repr.
        is_frozen = FROZEN in self.attrs
        lines = [
            "@define" if not is_frozen else "@frozen",
            f"class {self.name}:",
        ]
        field_lines = []
        for f in self.field_defs:
            def_str = ""
            has_repr = True
            if f.type == "bytes" or NO_REPR in f.attrs:
                has_repr = False

            needs_field = not has_repr

            if needs_field:
                field_args = []
                if f.default:
                    field_args.append(("default", f.default))
                if not has_repr:
                    field_args.append(("repr", "False"))
                field_str = ", ".join(f"{k}={v}" for k, v in field_args)
                def_str = f" = field({field_str})"
            elif f.default:
                def_str = f" = {f.default}"

            line = (
                f"    {f.name}: {f.type}{' | None' if f.is_optional else ''}{def_str}"
            )
            field_lines.append(line)

        if not self.field_defs:
            field_lines.append("    pass")
        script = lines + field_lines
        return script


@define
class Enum:
    name: str
    type: type[PythonEnum]
    fields: list[str]

    def render(self) -> Script:
        lines = [f"class {self.name}({self.type}):", *self.fields]
        return lines


@frozen
class Union:
    name: str
    members: Sequence[tuple[str, int | None]] = field(converter=tuple)

    def render(self) -> Script:
        member_names = []
        to_resolve = set()
        if len(self.members) == 1:
            # Special case when there is only one union member.
            member_name, union_val = self.members[0]
            to_resolve.add(member_name)
            if union_val is not None:
                # This has a union tag attached.
                member_name = f"Annotated[{member_name}, UnionVal({union_val})]"
            else:
                member_name = f"Annotated[{member_name}, UnionVal(1)]"
            member_names.append(member_name)
        else:
            for member_name, union_val in self.members:
                to_resolve.add(member_name)
                if union_val is not None:
                    # This has a union tag attached.
                    member_name = f"Annotated[{member_name}, UnionVal({union_val})]"
                member_names.append(member_name)
        member_string = " | ".join(member_names)
        return [f"{self.name} = {member_string}"]


@define
class Module:
    namespace: str | None
    path: Path
    imports: Imports
    importables: list[tuple[ImportableName, Union | Table | Enum | Attribute]]

    def render(self) -> str:
        body = ""
        # We potentially need to reorder since unions must refer to existing
        # names.
        tables_to_indices = {}
        for ix, item in enumerate(self.importables):
            if isinstance(item[1], Table):
                tables_to_indices[item[0]] = ix

        render_after: dict[Table, list[tuple[ImportableName, Union]]] = {}

        for ix, item in enumerate(list(self.importables)):
            if isinstance(item[1], Union):
                member_indices = [
                    tables_to_indices.get(n[0], -1) for n in item[1].members
                ]
                if member_indices:
                    largest_child_idx = max(member_indices)
                    if largest_child_idx > ix:
                        render_after.setdefault(
                            self.importables[largest_child_idx][1], []
                        ).append(item)
                        self.importables.remove(item)

        for item in self.importables:
            if isinstance(item[1], Table):
                body += "\n".join(item[1].render())
                post_unions = render_after.get(item[1], [])
                if post_unions:
                    body += "\n\n\n"
                for union in post_unions:
                    rendered = "\n".join(union[1].render())
                    body += rendered
                    body += "\n\n\n"
            elif isinstance(item[1], Enum):
                body += "\n".join(item[1].render())
                body += "\n"
            elif isinstance(item[1], Attribute):
                # Attributes don't get rendered.
                pass
            else:
                body += "\n".join(item[1].render())
            body += "\n\n\n"
        if not body.strip():
            # We don't render empty files.
            return ""

        imports = self.imports
        for _, item in self.importables:
            if isinstance(item, Table):
                imports = merge_imports(imports, item.imports)
        imports = {k: v for k, v in imports.items() if k != MISSING}

        script = (
            "from __future__ import annotations\n\n"
            + "\n".join(
                f"from {source} import {', '.join(sorted(targets))}"
                for source, targets in imports.items()
            )
            + "\n"
            + body
        )
        return script


class FlatbufferRenderer(Interpreter):
    """Transform FlatBuffer code into Python code.

    The rendering needs to be done in several phases, since imports
    need to be resolved after the initial parsing pass and types
    may need to be adjusted based on the actual types.
    """

    def module(self, tree: ParseTree, filename: Path) -> Module:
        namespace = None
        for t in tree.find_data("namespace"):
            namespace = str(t.children[0])
        importables: list[
            tuple[ImportableName, Script | Table | Enum | Attribute, Imports]
        ] = [
            self.visit(t)
            for t in tree.find_pred(
                lambda t: t.data in ("table", "enum", "union", "attribute")
            )
        ]
        imports = {}
        res = []

        local_names = {i[0] for i in importables}

        for name, script_or_table, imps in importables:
            if isinstance(script_or_table, Attribute):
                script_or_table = evolve(script_or_table, namespace=namespace)
            res.append((name, script_or_table))
            for k, v in imps.items():
                if k == MISSING:
                    for val in v:
                        if val in local_names:
                            continue
                        imports.setdefault(k, set()).add(val)
                else:
                    if k not in imports:
                        imports[k] = set()
                    imports[k] |= v

        return Module(namespace, filename, imports, res)

    def namespace(self, _: ParseTree) -> tuple[list[str], Imports]:
        return [], {}

    def attribute(self, tree: ParseTree) -> tuple[ImportableName, Attribute, Imports]:
        name = str(tree.children[0])
        return (name, Attribute(name), {})

    def union(self, tree: ParseTree) -> tuple[ImportableName, Union, Imports]:
        name = str(tree.children[0])
        member_names = []
        to_resolve = set()
        imports = {}
        members = tree.children[1:]
        if len(members) == 1:
            # Special case when there is only one union member.
            member = members[0]
            member_name = str(member.children[0]).split(".")[-1]
            to_resolve.add(member_name)
            imports["typing"] = {"Annotated"}
            imports["flattrs"] = {"UnionVal"}
            if len(member.children) > 1:
                # This has a union tag attached.
                member_names.append((member_name, int(member.children[1])))
            else:
                member_names.append((member_name, None))
        else:
            for member in tree.children[1:]:
                member_name = str(member.children[0]).split(".")[-1]
                to_resolve.add(member_name)
                if len(member.children) > 1:
                    # This has a union tag attached.
                    imports["typing"] = {"Annotated"}
                    imports["flattrs"] = {"UnionVal"}
                    member_names.append((member_name, int(member.children[1])))
                else:
                    member_names.append((member_name, None))
        return (
            name,
            Union(name, member_names),
            {MISSING: to_resolve} | imports,
        )

    def enum(self, tree: ParseTree) -> tuple[ImportableName, Enum, Imports]:
        enum_name = str(tree.children[0])
        type = ENUM_TYPE_MAP[tree.children[1]]
        if type == "IntEnum":
            imports = {"enum": {"IntEnum"}}
        else:
            imports = {"flattrs": {type}}
        fields = []
        next_val = 0
        for child in tree.children[2:]:
            name, val = self.visit(child)
            if val is None:
                val = next_val
                next_val += 1
            else:
                next_val = val + 1
            fields.append(f"    {map_python_keywords(name)} = {val}")
        return enum_name, Enum(enum_name, type, fields), imports

    def enum_field(self, tree: ParseTree) -> tuple[str, int | None]:
        val = None
        for enum_val in tree.find_data("enum_field_default"):
            val = int(enum_val.children[0])
        return tree.children[0], val

    def table(self, tree: ParseTree) -> tuple[ImportableName, Table, Imports]:
        name = str(tree.children[0])
        table_attributes = []
        for child in tree.children:
            if isinstance(child, Tree) and child.data == "attributes":
                table_attributes.extend(
                    Attribute(c, namespace=None if c in BUILTIN_ATTRS else MISSING)
                    for c in child.children
                )
        field_defs = [self.table_field(c) for c in tree.find_data("table_field")]
        imports = Map()
        for _, _, _, _, field_imports, _ in field_defs:
            imports = merge_imports(imports, field_imports)
        fields = []
        for f in field_defs:
            is_optional = Attribute("required", namespace=None) not in f[3]
            fields.append(
                Table.Field(
                    f[0],
                    f[1],
                    f[2],
                    is_optional,
                    f[5] or None,
                    f[3],
                )
            )

        table = Table(
            name,
            fields,
            imports,
            table_attributes,
        )
        return (
            name,
            table,
            imports,
        )

    def table_field(
        self, tree: ParseTree
    ) -> tuple[FieldName, str, str, list[Attribute], Imports, str]:
        imports = {}
        name = str(tree.children[0])
        full_type = tree.children[1]
        attributes = []
        for attr_tree in tree.find_data("attributes"):
            for c in attr_tree.children:
                attr_name = str(c)
                if attr_name in BUILTIN_ATTRS:
                    attr_namespace = None
                else:
                    attr_namespace = MISSING
                attributes.append(Attribute(attr_name, namespace=attr_namespace))
        is_scalar = False
        namespace_prefix = ""
        if full_type == "string":
            type = "str"
            default = ""
        elif isinstance(full_type, Token):
            # This is a name, so probably a table field.
            default = ""
            namespace_prefix, type = (
                full_type.rsplit(".", 1) if "." in full_type else ("", full_type)
            )
            imports[MISSING] = imports.get(MISSING, frozenset()) | {str(type)}
            for def_child in tree.find_data("table_field_default"):
                default = str(def_child.children[0])
        else:
            if full_type.data == "vector_type":
                inner_type = full_type.children[0]
                if inner_type.lower() in ("uint8", "ubyte"):
                    type = "bytes"
                else:
                    if inner_type.lower() in TYPE_MAP:
                        inner_type = TYPE_MAP[inner_type]
                        if inner_type not in ("int", "float", "bool", "str", "float"):
                            imports["flattrs"] = {inner_type}
                    else:
                        namespace_prefix, inner_type = (
                            inner_type.rsplit(".", 1)
                            if "." in inner_type
                            else ("", inner_type)
                        )
                        imports[MISSING] = imports.get(MISSING, frozenset()) | {
                            str(inner_type)
                        }
                    type = f"list[{inner_type}]"
                default = ""
            elif full_type.data == "string":
                pass
            else:
                # A scalar.
                attributes.append(Attribute("required", namespace=None))
                type = full_type.children[0]
                if type.lower() in TYPE_MAP:
                    is_scalar = True
                    type = TYPE_MAP[type]
                    if type not in ("int", "float", "bool", "str", "float"):
                        imports["flattrs"] = frozenset([type])

                default = ""
                for def_tree in tree.find_data("table_field_default"):
                    def_val = def_tree.children[0]
                    if type == "bool":
                        def_val = def_val.capitalize()
                    elif not is_scalar:
                        # This has to be an enum
                        def_val = f"{type}.{def_val}"
                    default = def_val

        return name, type, default, attributes, imports, namespace_prefix

    def table_field_default(self):
        pass


def parse_module(module_tree: Tree, file: Path) -> Module:
    r = FlatbufferRenderer()
    parsed_module = r.module(module_tree, file)
    return parsed_module


def render_directory(
    input: Path, output: Path, gen_namespace_imports: bool = False
) -> None:
    """
    Render a directory of Flatbuffers into a directory of Python.

    """
    # First we parse, then we resolve imports, then we write.
    per_namespace: dict[str | None, list[Module]] = {}
    importables_to_module: dict[str, Module] = {}
    namespace_to_attributes: dict[str, set[str]] = {}
    enums = set()
    modules: list[Module] = []
    for file in input.rglob("*.fbs"):
        rel_path = file.relative_to(input).with_suffix(".py")
        try:
            pm = parse_module(parser.parse(file.read_text()), rel_path)
        except Exception as exc:
            print(f"While parsing {file}: {exc}")
            print_exception(exc)
        modules.append(pm)
        for name, importable in pm.importables:
            importables_to_module[name] = pm
            if isinstance(importable, Enum):
                enums.add(importable.name)
            elif isinstance(importable, Attribute):
                namespace_to_attributes.setdefault(pm.namespace or "", set()).add(
                    importable.name
                )
        per_namespace.setdefault(pm.namespace, []).append(pm)

    # We can resolve attribute namespaces now.
    # At this point, we know which importables are enums.
    # Tables referencing those need to have their types adjusted.
    # For each table, add defaults as much as possible,
    # starting from backwards. This is so fields added at the end
    # keep backwards comp.
    for module in modules:
        module.importables = [
            (
                name,
                i.adjust_attribute_namespaces(namespace_to_attributes)
                .adjust_enums(enums)
                .adjust_defaults()
                .adjust_imports()
                if isinstance(i, Table)
                else i,
            )
            for name, i in module.importables
        ]

    # Now we write the modules to disk.
    output.mkdir(exist_ok=True, parents=True)
    (output / "__init__.py").write_text("")
    for pm_list in per_namespace.values():
        for pm in pm_list:
            pm_path = pm.path
            missing_imports = pm.imports.pop(MISSING, set())
            for importable in missing_imports:
                module = importables_to_module[importable]

                if module.path.parent == pm_path.parent:
                    rel_path = f".{module.path.stem}"
                else:
                    pm_parents = set(pm_path.parents)
                    target_parents = set(module.path.parents)
                    common_parent = sorted(
                        pm_parents & target_parents, key=lambda p: -len(str(p))
                    )[0]
                    num_dots = len(pm_path.relative_to(common_parent).parent.parts) + 1

                    target_rel_path = module.path.relative_to(common_parent)
                    fname = target_rel_path.stem
                    if target_rel_path.parent == Path("."):
                        rel_path = f"{'.' * num_dots}{fname}"
                    else:
                        rel_path = f"{'.' * num_dots}{str(target_rel_path.parent).replace('/', '.')}.{fname}"
                pm.imports.setdefault(str(rel_path), set()).add(importable)
            try:
                target_file = output / pm.path
                rendered = pm.render()
                if not rendered:
                    continue
                target_file.parent.mkdir(exist_ok=True, parents=True)
                (target_file.parent / "__init__.py").write_text("")
                target_file.write_text(rendered)
            except Exception as exc:
                print(f"While parsing or rendering {file}: {exc}")
                print_exception(exc)
    if gen_namespace_imports:
        for namespace, pm_list in per_namespace.items():
            importables: list[tuple[str, Path]] = []
            for module in pm_list:
                importables.extend(
                    [
                        (i[0], module.path)
                        for i in module.importables
                        if not isinstance(i[1], Attribute)
                    ]
                )
            if not importables:
                continue
            namespace_dir = Path(
                namespace.replace(".", "/") if namespace is not None else "."
            )
            namespace_fn = namespace_dir / "__init__.py"
            (output / namespace_dir).mkdir(parents=True, exist_ok=True)
            (output / namespace_fn).write_text(
                render_exports(namespace_fn, importables)
            )


def make_import_path(importing_from: Path, import_target: Path) -> str:
    if importing_from.parent == import_target.parent:
        rel_path = f".{import_target.stem}"
    else:
        active_parents = set(importing_from.parents)
        target_parents = set(import_target.parents)
        common_parent = sorted(
            active_parents & target_parents, key=lambda p: -len(str(p))
        )[0]
        num_dots = len(importing_from.relative_to(common_parent).parent.parts) + 1

        target_rel_path = import_target.relative_to(common_parent)
        fname = target_rel_path.stem
        if target_rel_path.parent == Path("."):
            rel_path = f"{'.' * num_dots}{fname}"
        else:
            rel_path = f"{'.' * num_dots}{str(target_rel_path.parent).replace('/', '.')}.{fname}"
    return rel_path


def render_exports(importing_from: Path, importables: list[tuple[str, Path]]) -> str:
    return (
        "\n".join(
            f"from {make_import_path(importing_from, p)} import {i}"
            for i, p in importables
        )
        + "\n\n__all__ = ["
        + "\n".join(f'"{i[0]}",' for i in sorted(importables))
        + "]\n"
    )


def render(input: Path, output: Path, gen_namespace_exports: bool = False) -> None:
    if input.is_file():
        output.parent.mkdir(exist_ok=True, parents=True)
        output.write_text(parse_module(parser.parse(input.read_text()), output))
    else:
        render_directory(input, output, gen_namespace_exports)


def merge_imports(a: Imports, b: Imports) -> Imports:
    res: Imports = {}
    for k, v in a.items():
        res[k] = frozenset(res.get(k, frozenset()) | v)
    for k, v in b.items():
        res[k] = frozenset(res.get(k, frozenset()) | v)
    return Map(res)


def map_python_keywords(name: str) -> str:
    """If an identifier is a Python keyword, we adjust it by adding an underscore.

    Like flatc.
    """
    return f"{name}_" if iskeyword(name) else name


TYPE_MAP: Final = {
    "string": "str",
    "bool": "bool",
    "uint8": "Uint8",
    "ubyte": "Uint8",
    "uint16": "Uint16",
    "ushort": "Uint16",
    "uint32": "Uint32",
    "uint": "Uint32",
    "uint64": "Uint64",
    "ulong": "Uint64",
    "int8": "Int8",
    "byte": "Int8",
    "int16": "Int16",
    "short": "Int16",
    "int32": "Int32",
    "int": "Int32",
    "int64": "int",
    "long": "int",
    "float": "Float32",
    "float32": "Float32",
    "double": "float",
    "float64": "float",
}
ENUM_TYPE_MAP: Final = {
    "uint8": "Uint8Enum",
    "ubyte": "Uint8Enum",
    "uint16": "Uint16Enum",
    "ushort": "Uint16Enum",
    "uint32": "Uint32Enum",
    "uint": "Uint32Enum",
    "uint64": "Uint64Enum",
    "ulong": "Uint64Enum",
    "int8": "Int8Enum",
    "byte": "Int8Enum",
    "int16": "Int16Enum",
    "short": "Int16Enum",
    "int32": "Int32Enum",
    "int": "Int32Enum",
    "int64": "IntEnum",
    "long": "IntEnum",
}



"""Types used internally."""
from typing import Any, Literal, TypeAlias

from attrs import NOTHING, AttrsInstance

FieldName: TypeAlias = str
SlotNumber: TypeAlias = int
MaybeDefault: TypeAlias = Any | Literal[NOTHING]
ScalarType: TypeAlias = Literal[
    "Bool", "Uint8", "Uint16", "Uint32", "Float32", "float", "Float64", "double"
]
PythonScalarType: TypeAlias = type[int] | type[bool] | type[float]
Optionality: TypeAlias = bool
UnionMapping: TypeAlias = dict[int, type[AttrsInstance]]



"""Misc tests for coverage."""
from enum import Enum

from attrs import define
from pytest import raises

from flattrs import dumps, make_dumps, make_loads

from .flattrs.models.common import Common1


def test_unsupported_enums() -> None:
    """Unsupported enums throw errors."""

    class TestEnum(str, Enum):
        TEST = "test"

    with raises(TypeError):

        @define
        class Test:
            test: TestEnum

        dumps(Test(TestEnum.TEST))

    with raises(TypeError):

        @define
        class Test:
            test: TestEnum | None

        dumps(Test(TestEnum.TEST))

    with raises(TypeError):

        @define
        class Test:
            test: list[TestEnum]

        dumps(Test([TestEnum.TEST]))

    with raises(TypeError):

        @define
        class Test:
            test: list[TestEnum] | None

        dumps(Test(TestEnum.TEST))


def test_make_fns() -> None:
    """flattrs.make_* work properly."""
    dumps = make_dumps(Common1)
    loads = make_loads(Common1)
    assert loads(dumps(i := Common1("id", 1, 2))) == i



from enum import IntEnum
from typing import Annotated, Literal, TypeAlias

from attrs import frozen

try:
    from enum import ReprEnum as Enum
except ImportError:
    from enum import Enum


@frozen
class ScalarMarker:
    signed: bool
    width: Literal[8, 16, 32, 64]


Uint8 = Annotated[int, ScalarMarker(False, 8)]
Uint16 = Annotated[int, ScalarMarker(False, 16)]
Uint32 = Annotated[int, ScalarMarker(False, 32)]
Uint64 = Annotated[int, ScalarMarker(False, 64)]
Int8 = Annotated[int, ScalarMarker(True, 8)]
Int16 = Annotated[int, ScalarMarker(True, 16)]
Int32 = Annotated[int, ScalarMarker(True, 32)]
Int64: TypeAlias = int
Float = Annotated[float, ScalarMarker(True, 32)]
Float32: TypeAlias = Float
Float64: TypeAlias = float


class Uint8Enum(IntEnum):
    """An 8-bit Flatbuffer enum."""


UbyteEnum: TypeAlias = Uint8Enum


class Uint16Enum(IntEnum):
    """A 16-bit Flatbuffer enum."""


class Uint32Enum(IntEnum):
    """A 32-bit Flatbuffer enum."""


class Uint64Enum(IntEnum):
    """A 64-bit Flatbuffer enum."""


class Int8Enum(Int8, Enum):
    """An 8-bit Flatbuffer signed enum."""


ByteEnum: TypeAlias = Int8Enum


class Int16Enum(IntEnum):
    """A 16-bit Flatbuffer signed enum."""


class Int32Enum(IntEnum):
    """A 32-bit Flatbuffer signed enum."""


Int64Enum: TypeAlias = IntEnum


class UnionVal(int):
    """Annotate a union member with this to set the union type value."""

    pass



from typing import Final

from .converters import Converter
from .types import (
    Float,
    Float32,
    Float64,
    Int8,
    Int8Enum,
    Int16,
    Int16Enum,
    Int32,
    Int32Enum,
    Int64,
    Int64Enum,
    Uint8,
    Uint8Enum,
    Uint16,
    Uint16Enum,
    Uint32,
    Uint32Enum,
    Uint64,
    Uint64Enum,
    UnionVal,
)

global_converter = Converter()
dumps: Final = global_converter.dumps
loads: Final = global_converter.loads
make_dumps: Final = global_converter.make_dumps
make_loads: Final = global_converter.make_loads

__all__ = [
    "dumps",
    "Float",
    "Float32",
    "Float64",
    "Int16",
    "Int16Enum",
    "Int32",
    "Int32Enum",
    "Int64",
    "Int64Enum",
    "Int8",
    "Int8Enum",
    "loads",
    "make_dumps",
    "make_loads",
    "Uint16",
    "Uint16Enum",
    "Uint32",
    "Uint32Enum",
    "Uint64",
    "Uint64Enum",
    "Uint8",
    "Uint8Enum",
    "UnionVal",
]



from pathlib import Path
from subprocess import run

from flattrs.modgen.renderer import render


def assert_dirs_identical(dir_a: Path, dir_b: Path) -> None:
    assert dir_a.is_dir()
    assert dir_b.is_dir()
    list_a = [
        p.relative_to(dir_a) for p in dir_a.glob("*") if p.name not in ("__pycache__")
    ]
    list_b = [
        p.relative_to(dir_b) for p in dir_b.glob("*") if p.name not in ("__pycache__")
    ]
    assert list_a == list_b

    for elem_a, elem_b in zip(list_a, list_b):
        file_a = dir_a / elem_a
        file_b = dir_b / elem_b
        if file_a.is_file():
            assert file_b.is_file()
            assert file_a.read_bytes() == file_b.read_bytes()
        else:
            assert not file_b.is_file()
            assert_dirs_identical(file_a, file_b)


def test_compare_to_manual(tmp_path: Path) -> None:
    """The manually inspected models are in ./flattrs/models_manual.

    These should be identical to generated models.
    """
    dest_dir = tmp_path / "models"
    run(
        [
            "python",
            "-m",
            "coverage",
            "run",
            "-m",
            "flattrs.modgen",
            str((Path(__file__).parent / "flatbufferdefs")),
            str(dest_dir),
            "--gen-namespace-exports",
        ],
        check=True,
    )
    run(["isort", str(dest_dir)], check=True)
    run(["black", str(dest_dir)], check=True)
    assert_dirs_identical(Path(__file__).parent / "flattrs" / "models_manual", dest_dir)



from types import GenericAlias, UnionType
from typing import Any, TypeVar, Union, _AnnotatedAlias, _GenericAlias, get_args

from attr._make import _obj_setattr
from attrs import fields

from ._consts import NoneType
from ._types import Optionality

T = TypeVar("T")


def is_subclass(typ, supertype) -> bool:
    """A safe version of `is_subclass`."""
    try:
        return issubclass(typ, supertype)
    except TypeError:
        return False


def is_generic_subclass(t, s):
    return (isinstance(t, _GenericAlias) and t.__origin__ is s) or (
        isinstance(t, GenericAlias) and t.__origin__ is s
    )


def get_annotation_and_base(typ, annotation_cls: type[T]) -> tuple[T, Any] | None:
    if typ.__class__ is _AnnotatedAlias:
        args = get_args(typ)
        for arg in args[1:]:
            if isinstance(arg, annotation_cls):
                return arg, args[0]
    return None


def get_optional_arg(type) -> Any | None:
    """Is this annotation an `Optional` (union of something and None)?"""
    if getattr(type, "__origin__", None) is Union or type.__class__ is UnionType:
        union_args = type.__args__
        if len(union_args) == 2 and NoneType in union_args:
            return [a for a in union_args if a is not NoneType][0]

    return None


def get_union_args(type) -> tuple[type, ...] | None:
    """Is this annotation a Union?"""
    if getattr(type, "__origin__", None) is Union or type.__class__ is UnionType:
        return type.__args__

    return None


def get_list_args(type) -> tuple[Any, Optionality] | None:
    """Is this type annotation a list of something?"""
    optional = False
    if (o := get_optional_arg(type)) is not None:
        optional = True
        type = o
    if is_generic_subclass(type, list):
        arg = get_args(type)[0]
        return arg, optional
    return None


def resolve_types(cls, globalns=None, localns=None, attribs=None, include_extras=True):
    # Since calling get_type_hints is expensive we cache whether we've
    # done it already.
    if getattr(cls, "__attrs_types_resolved__", None) != cls:
        import typing

        kwargs = {
            "globalns": globalns,
            "localns": localns,
            "include_extras": include_extras,
        }

        hints = typing.get_type_hints(cls, **kwargs)
        for field in fields(cls) if attribs is None else attribs:
            if field.name in hints:
                # Since fields have been frozen we must work around it.
                _obj_setattr(field, "type", hints[field.name])
        # We store the class we resolved so that subclasses know they haven't
        # been resolved.
        cls.__attrs_types_resolved__ = cls

    # Return the class so you can use it as a decorator too.
    return cls



from platform import python_implementation

from setuptools import find_packages, setup

if python_implementation() != "PyPy":
    from Cython.Build import cythonize

    ext_modules = cythonize("src/flattrs/cflattrs/*.pyx", annotate=True)
else:
    ext_modules = []

setup(
    name="flattrs",
    long_description="Flatbuffers support for Python",
    long_description_content_type="text/x-rst",
    version="23.1.0.b7",
    install_requires=[
        "attrs",
        "flatbuffers==23.1.4",
        "click",
        "lark >= 1.1.5",
        "immutables",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    ext_modules=ext_modules,
    zip_safe=False,
    extras_require={
        "dev": [
            "pytest",
            "hypothesis",
            "cython==0.29.32",
            "coverage",
            "cattrs",
            "nox",
            "numpy",
            "black==23.1.0",
            "isort",
        ]
    },
    package_data={"": ["*.pxd", "*.pyx", "py.typed"]},
    license="MIT license",
    entry_points={"console_scripts": ["flattrs=flattrs.modgen.__main__:main"]},
)



import hashlib
import linecache
import struct
from functools import cache
from typing import Any, Callable, Final, TypeAlias, TypeVar

from attrs import NOTHING, AttrsInstance, Factory, define, fields
from flatbuffers.number_types import (
    BoolFlags,
    Float32Flags,
    Float64Flags,
    Int8Flags,
    Int16Flags,
    Int32Flags,
    Int64Flags,
    Uint8Flags,
    Uint16Flags,
    Uint32Flags,
    Uint64Flags,
)
from flatbuffers.table import Table

from ._analysis import FlatbufferTable, analyze
from ._consts import (
    SCALAR_TYPE_TO_DEFAULT,
    SCALAR_TYPE_TO_PYTHON,
    SCALAR_TYPE_TO_WIDTH,
    NoneType,
)
from ._types import ScalarType
from .types import UnionVal
from .typing import get_annotation_and_base

try:
    from .cflattrs.builder import Builder
except ImportError:  # NOQA
    from flatbuffers.builder import Builder

T = TypeVar("T", bound=AttrsInstance)
AddToBuilder: TypeAlias = Callable[[T, Builder, dict[int, int], dict[int, int]], int]


@define
class Converter:
    """A converter for loading and dumping Flatbuffers."""

    _dumps_cache: Callable = Factory(
        lambda self: cache(lambda cls: self.make_dumps(cls)), takes_self=True
    )
    _to_fb_cache: Callable[[type[T]], AddToBuilder] = Factory(
        lambda self: cache(
            lambda cls: _make_add_to_builder_fn(analyze(cls), self._to_fb_cache)
        ),
        takes_self=True,
    )
    _from_fb_cache: Callable[[type[T]], Callable] = Factory(
        lambda self: cache(lambda cls: self._make_from_fb(cls)), takes_self=True
    )
    _loads_cache: Callable = Factory(
        lambda self: cache(lambda cls: self.make_loads(cls)), takes_self=True
    )

    def loads(self, payload: bytes, cl: type[T]) -> T:
        return self._loads_cache(cl)(payload)

    def dumps(self, model: AttrsInstance, builder: Builder | None = None) -> bytes:
        return self._dumps_cache(model.__class__)(model, builder)

    def _make_from_fb(self, cl: type[AttrsInstance]) -> Callable:
        """Generate all necessary functions for a class to work with Flatbuffers."""
        return make_from_fb_fn(analyze(cl), self._from_fb_cache)

    def make_loads(self, cl: type[T]) -> Callable[[bytes], T]:
        """Prepare a loading function for a model in advance."""
        from_fb = self._from_fb_cache(cl)
        start_struct = struct.Struct("<I")

        def loads(data: bytes, _from_fb=from_fb, _start_struct=start_struct) -> T:
            start_offset = _start_struct.unpack_from(data, 0)[0]
            return _from_fb(Table(data, start_offset))

        return loads

    def make_dumps(self, cl: type[T]) -> Callable[[T, Builder | None], bytes]:
        """Prepare a dumping function for a model in advance."""

        fb_cl = analyze(cl)

        add_to_builder = _make_add_to_builder_fn(fb_cl, self._to_fb_cache)

        def dumps(model: T, builder: Builder | None = None) -> bytes:
            builder = Builder(10000) if builder is None else builder
            offset = add_to_builder(model, builder, {}, {})
            builder.Finish(offset)
            return builder.Output()

        return dumps


def _make_add_to_builder_fn(
    cl: FlatbufferTable,
    hook_factory: Callable[[type[AttrsInstance]], AddToBuilder[Any]],
) -> AddToBuilder[Any]:
    name = cl.cl.__name__
    globs = {}
    lines = []
    lines.append("def __fb_add_to_builder__(self, builder, strs, nodes):")

    for field, slot_idx, is_optional in cl.strings:
        lines.append(f"    __fb_self_{field} = self.{field}")
        if is_optional:
            lines.append(
                f"    if __fb_self_{field} is not None and __fb_self_{field} not in strs:"
            )
        else:
            lines.append(f"    if __fb_self_{field} not in strs:")

        lines.append(
            f"        strs[__fb_self_{field}] = builder.CreateString(__fb_self_{field})"
        )

    for field, slot_idx, is_optional in cl.byte_fields:
        lines.append(f"    __fb_self_{field} = self.{field}")
        infix = ""
        if is_optional:
            infix = f"__fb_self_{field} is not None and "
        lines.append(f"    if {infix}id(__fb_self_{field}) not in nodes:")
        lines.append(
            f"        nodes[id(__fb_self_{field})] = builder.CreateByteVector(__fb_self_{field})"
        )

    for field, _, is_optional in cl.lists_of_strings:
        lines.append(f"    __fb_self_{field} = self.{field}")
        i = ""
        if is_optional:
            i = "    "
            lines.append(f"    if __fb_self_{field} is not None:")

        lines.append(f"{i}    __fb_self_{field}_offsets = []")
        lines.append(f"{i}    for e in __fb_self_{field}:")
        lines.append(f"{i}        if e in strs:")
        lines.append(f"{i}            __fb_self_{field}_offsets.append(strs[e])")
        lines.append(f"{i}        else:")
        lines.append(f"{i}            offset = builder.CreateString(e)")
        lines.append(f"{i}            strs[e] = offset")
        lines.append(f"{i}            __fb_self_{field}_offsets.append(offset)")
        lines.append(f"{i}    builder.StartVector(4, len(__fb_self_{field}), 4)")
        lines.append(f"{i}    for o in reversed(__fb_self_{field}_offsets):")
        lines.append(f"{i}        builder.PrependUOffsetTRelative(o)")
        lines.append(f"{i}    __fb_self_{field}_offset = builder.EndVector()")

    for field, _, _, fb_number_type, is_optional in (
        cl.lists_of_scalars + cl.lists_of_enums
    ):
        prepend = SCALAR_TYPE_TO_PREPEND[fb_number_type]
        w = SCALAR_TYPE_TO_WIDTH[fb_number_type]

        lines.append(f"    __fb_self_{field} = self.{field}")
        indent = ""
        if is_optional:
            lines.append(f"    if __fb_self_{field} is not None:")
            indent = "  "
        lines.append(f"    {indent}builder.StartVector({w}, len(self.{field}), {w})")
        lines.append(f"    {indent}for o in reversed(__fb_self_{field}):")
        lines.append(f"    {indent}    builder.{prepend}(o)")
        lines.append(f"    {indent}__fb_self_{field}_offset = builder.EndVector()")

    for field, type, slot_idx, is_optional in cl.lists_of_tables:
        globs[f"_{field}_add_to_builder"] = hook_factory(type)
        lines.append(f"    __fb_self_{field} = self.{field}")

        indent = ""
        if is_optional:
            lines.append(f"    if __fb_self_{field} is not None:")
            indent = "    "

        lines.append(
            f"    {indent}offsets = [_{field}_add_to_builder(t, builder, strs, nodes) for t in __fb_self_{field}]"
        )
        lines.append(f"    {indent}builder.StartVector(4, len(offsets), 4)")
        lines.append(
            f"    {indent}for o in reversed(offsets): builder.PrependUOffsetTRelative(o)"
        )
        lines.append(f"    {indent}nodes[id(__fb_self_{field})] = builder.EndVector()")

    for field, type, _, is_optional in cl.tables:
        globs[f"_{field}_add_to_builder"] = hook_factory(type)
        prefix = ""
        if is_optional:
            prefix = f"if self.{field} is not None: "
        lines.append(
            f"    {prefix}nodes[id(self.{field})] = _{field}_add_to_builder(self.{field}, builder, strs, nodes)"
        )

    for field, union_members, _, _ in cl.unions:
        prefix = ""
        is_optional = NoneType in union_members
        if is_optional:
            prefix = f"if self.{field} is not None: "
        union_hooks = {}
        for type in union_members:
            if type is NoneType:
                continue
            # It's possible this is an annotation
            annot_and_base = get_annotation_and_base(type, UnionVal)
            if annot_and_base is not None:
                base = annot_and_base[1]
                union_hooks[base] = hook_factory(base)
            else:
                union_hooks[type] = hook_factory(type)
        globs[f"_{field}_hooks"] = union_hooks
        lines.append(
            f"    {prefix}nodes[id(self.{field})] = _{field}_hooks[self.{field}.__class__](self.{field}, builder, strs, nodes)"
        )

    # The actual object starts here.
    lines.append(f"    builder.StartObject({cl.num_slots})")

    for field, slot_idx, is_optional in cl.strings:
        indent = ""
        if is_optional:
            lines.append(f"    if __fb_self_{field} is not None:")
            indent = "    "
        lines.append(
            f"    {indent}builder.PrependUOffsetTRelativeSlot({slot_idx}, strs[__fb_self_{field}], 0)"
        )

    for field, slot_idx, is_optional in cl.byte_fields:
        prefix = ""
        if is_optional:
            prefix = f"if __fb_self_{field} is not None: "
        lines.append(
            f"    {prefix}builder.PrependUOffsetTRelativeSlot({slot_idx}, nodes[id(__fb_self_{field})], 0)"
        )

    for field, slot_idx, is_optional in cl.lists_of_strings:
        prefix = ""
        if is_optional:
            prefix = f"if __fb_self_{field} is not None: "
        lines.append(
            f"    {prefix}builder.PrependUOffsetTRelativeSlot({slot_idx}, __fb_self_{field}_offset, 0)"
        )

    for field, _, slot_idx, is_optional in cl.tables:
        indent = ""
        if is_optional:
            lines.append(f"    if self.{field} is not None:")
            indent = "    "
        lines.append(
            f"    {indent}builder.PrependUOffsetTRelativeSlot({slot_idx}, nodes[id(self.{field})], 0)"
        )

    for field, _, slot_idx, is_optional in cl.lists_of_tables:
        lines.append(f"    __fb_self_{field} = self.{field}")
        if is_optional:
            indent = f"if __fb_self_{field} is not None: "
        else:
            indent = ""
        lines.append(
            f"    {indent}builder.PrependUOffsetTRelativeSlot({slot_idx}, nodes[id(__fb_self_{field})], 0)"
        )

    for field, slot_idx, _, _, is_optional in cl.lists_of_scalars + cl.lists_of_enums:
        if is_optional:
            prefix = f"if __fb_self_{field} is not None: "
        else:
            prefix = ""
        lines.append(
            f"    {prefix}builder.PrependUOffsetTRelativeSlot({slot_idx}, __fb_self_{field}_offset, 0)"
        )

    for field, fb_type, slot_idx, default in cl.inlines + cl.enums:
        if default is NOTHING:
            default = SCALAR_TYPE_TO_DEFAULT[fb_type]
        lines.append(
            f"    builder.Prepend{fb_type}Slot({slot_idx}, self.{field}, {default})"
        )

    for field, union_types, union_mapping, slot_idx in cl.unions:
        union_dict_name = f"_{field}_union_types"
        # Unions might be prefixed by a namespace string, depending on how
        # they're defined.
        # stripped_union_dict = {k.split("_")[-1]: v for k, v in fb_enum.__dict__.items()}

        union_dict = {t: ix for ix, t in union_mapping.items() if t is not NoneType}
        # Flatbuffer unions always have a special member, NONE, signifying
        # a missing value. If None is part of union_types, we handle it
        # specially.
        if NoneType in union_types:
            union_dict[NoneType] = 0

        globs[union_dict_name] = union_dict
        lines.append(
            f"    builder.PrependUint8Slot({slot_idx}, {union_dict_name}[self.{field}.__class__], 0)"
        )
        indent = ""
        if NoneType in union_types:
            lines.append(f"    if self.{field} is not None:")
            indent = "    "
        lines.append(
            f"    {indent}builder.PrependUOffsetTRelativeSlot({slot_idx+1}, nodes[id(self.{field})], 0)"
        )

    lines.append("    return builder.EndObject()")
    lines.append("")
    sha1 = hashlib.sha1()
    sha1.update(name.encode("utf-8"))
    unique_filename = "<FB add_to_builder for %s, %s>" % (
        name,
        sha1.hexdigest(),
    )
    script = "\n".join(lines)
    eval(compile(script, unique_filename, "exec"), globs)

    linecache.cache[unique_filename] = (
        len(script),
        None,
        script.splitlines(True),
        unique_filename,
    )

    return globs["__fb_add_to_builder__"]


def make_from_fb_fn(
    cl: FlatbufferTable, hook_factory: Callable[[type[AttrsInstance]], Callable]
) -> Callable:
    """Compile a function to init an attrs model from a FB model, flattrs-style.

    The input is assumed to be an instance of `flatbuffers.Table`.
    """
    name = cl.cl.__name__
    globs = {"_Table": Table, "cls": cl.cl}
    lines = []
    inst_lines = []  # Instantiation lines
    byte_names = {f[0]: f for f in cl.byte_fields}
    string_names = {s[0]: s for s in cl.strings}
    list_table_fields = {t[0]: t for t in cl.lists_of_tables}
    union_field_names = {t[0]: t for t in cl.unions}
    lists_of_scalar_names = {t[0]: t for t in cl.lists_of_scalars + cl.lists_of_enums}
    lists_of_strings_names = {t[0]: t for t in cl.lists_of_strings}
    inline_names = {t[0]: t for t in cl.inlines}
    enum_names = {e[0]: e for e in cl.enums}
    table_names = {t[0]: t for t in cl.tables}

    lines.append("def __fb_from_fb__(tab):")

    inst_lines.append("    return cls(")
    for field in fields(cl.cl):
        fname = field.name
        if fname in string_names:
            string_def = string_names[fname]
            slot_idx = string_def[1]
            if string_def[2]:
                inst_lines.append(
                    f"        tab.String(tab.Pos + o).decode() if (o := tab.Offset({4 + slot_idx * 2})) != 0 else None,"
                )
            else:
                inst_lines.append(
                    f"        tab.String(tab.Pos + tab.Offset({4 + slot_idx * 2})).decode(),"
                )
        elif fname in byte_names:
            byte_def = byte_names[fname]
            slot_idx = byte_def[1]
            if byte_def[2]:
                inst_lines.append(
                    f"        tab.String(tab.Pos + o) if (o := tab.Offset({4 + slot_idx * 2})) != 0 else None,"
                )
            else:
                inst_lines.append(
                    f"        tab.String(tab.Pos + tab.Offset({4 + slot_idx * 2})),"
                )
        elif fname in enum_names:
            _, fb_type, slot_idx, default = enum_names[fname]
            globs[f"_{fname}_type"] = field.type
            globs[f"_{fname}_flags"] = SCALAR_TYPES_TO_FLAGS[fb_type]
            inst_lines.append(
                f"        _{fname}_type(tab.Get(_{fname}_flags, tab.Pos + o) if (o := tab.Offset({4 + slot_idx * 2})) != 0 else {default}),"
            )
        elif fname in table_names:
            table_def = table_names[fname]
            slot_idx = table_def[2]
            globs[f"_{fname}_from_fb"] = hook_factory(table_def[1])
            lines.append(f"    _{fname}_offset = tab.Offset({4 + slot_idx * 2})")
            suffix = ""
            if table_def[3]:
                suffix = f" if _{fname}_offset != 0 else None"
            inst_lines.append(
                f"        _{fname}_from_fb(tab.__class__(tab.Bytes, tab.Indirect(_{fname}_offset + tab.Pos))){suffix},"
            )
        elif fname in list_table_fields:
            _, type, slot_idx, is_optional = list_table_fields[fname]
            hook = hook_factory(type)
            globs[f"_{fname}_from_fb"] = hook
            lines.append(f"    _{fname}_offset = tab.Offset({4 + slot_idx * 2})")
            lines.append(
                f"    _{fname}_vector = tab.Vector(_{fname}_offset) if _{fname}_offset != 0 else None"
            )
            for_ = f"for i in range(tab.VectorLen(_{fname}_offset) if _{fname}_offset != 0 else 0)"
            line = f"        [_{fname}_from_fb(tab.__class__(tab.Bytes, tab.Indirect(_{fname}_vector + 4 * i))) {for_}]"

            if is_optional:
                line = f"{line} if _{fname}_offset != 0 else None"
            inst_lines.append(f"{line},")
        elif fname in lists_of_strings_names:
            list_def = lists_of_strings_names[fname]
            is_optional = list_def[2]
            slot_idx = list_def[1]
            lines.append(f"    _{fname}_offset = tab.Offset({4 + slot_idx * 2})")
            lines.append(
                f"    _{fname}_vector = tab.Vector(_{fname}_offset) if _{fname}_offset != 0 else None"
            )
            for_ = f"for i in range(tab.VectorLen(_{fname}_offset))"
            line = f"        [tab.String(_{fname}_vector + i * 4).decode() {for_}]"
            if is_optional:
                # We adjust the previous line a little
                line = f"{line} if _{fname}_offset != 0 else None"
            inst_lines.append(f"{line},")
        elif fname in inline_names:
            inline = inline_names[fname]
            slot_idx = inline[2]
            globs[f"_{fname}_type"] = SCALAR_TYPE_TO_PYTHON[inline[1]]
            globs[f"_{fname}_flags"] = SCALAR_TYPES_TO_FLAGS[inline[1]]
            if inline[3] is not NOTHING:
                globs[f"_{fname}_default"] = inline[3]
                default = f"_{fname}_default"
            else:
                default = SCALAR_TYPE_TO_DEFAULT[inline[1]]
            inst_lines.append(
                f"        _{fname}_type(tab.Get(_{fname}_flags, tab.Pos + o)) if (o := tab.Offset({4 + slot_idx * 2})) != 0 else {default},"
            )
        elif fname in union_field_names:
            # We prepare a dictionary to select the proper class at runtime.
            # Then we just grab the proper type and instantiate it.
            _, union_types, union_mapping, slot_idx = union_field_names[fname]
            dn = f"_{fname}_union_dict"
            union_resolution_dict = {}

            for ix, union_type in union_mapping.items():
                # Unions might be prefixed by a namespace string, depending on how
                # they're defined.
                attr_model = union_type
                hook = hook_factory(attr_model)

                def _load_from_content(table: Table, o, hook=hook):
                    obj = Table(bytearray(), 0)
                    table.Union(obj, o)
                    return hook(obj)

                union_resolution_dict[ix] = _load_from_content

            globs[dn] = union_resolution_dict
            globs[f"_{fname}_flags"] = Uint8Flags
            lines.append(f"    _{fname}_offset = tab.Offset({4 + slot_idx * 2})")
            if NoneType in union_types:
                inst_lines.append(
                    f"        {dn}[tab.Get(_{fname}_flags, _{fname}_offset + tab.Pos)](tab, tab.Offset({6 + slot_idx * 2})) if _{fname}_offset != 0 else None,"
                )
            else:
                inst_lines.append(
                    f"        {dn}[tab.Get(_{fname}_flags, _{fname}_offset + tab.Pos) if _{fname}_offset != 0 else None](tab, tab.Offset({6 + slot_idx * 2})),"
                )
        elif fname in lists_of_scalar_names:
            list_def = lists_of_scalar_names[fname]
            flags = SCALAR_TYPES_TO_FLAGS[list_def[3]]
            globs[f"_{fname}_flags"] = flags
            globs[f"_{fname}_type"] = list_def[2]
            slot_idx = list_def[1]
            lines.append(f"    _{fname}_offset = tab.Offset({4 + slot_idx * 2})")
            lines.append(
                f"    _{fname}_vector = tab.Vector(_{fname}_offset) if _{fname}_offset != 0 else None"
            )
            for_ = f"for i in range(tab.VectorLen(_{fname}_offset) if _{fname}_offset != 0 else 0)"
            i = "i" if flags.bytewidth == 1 else f"i * {flags.bytewidth}"
            line = f"        [_{fname}_type(tab.Get(_{fname}_flags, _{fname}_vector + {i})) {for_}]"
            if list_def[4]:
                # We adjust the previous line a little
                line = f"{line} if _{fname}_offset != 0 else None"
            inst_lines.append(line + ", ")
        else:
            raise ValueError(f"Can't handle {fname} (type {field.type}).")

    inst_lines.append("    )")

    sha1 = hashlib.sha1()
    sha1.update(name.encode("utf-8"))
    unique_filename = "<FB from_fb for %s, %s>" % (name, sha1.hexdigest())
    script = "\n".join(lines + inst_lines)
    eval(compile(script, unique_filename, "exec"), globs)

    linecache.cache[unique_filename] = (
        len(script),
        None,
        script.splitlines(True),
        unique_filename,
    )

    return globs["__fb_from_fb__"]


SCALAR_TYPE_TO_PREPEND: Final[dict[ScalarType, str]] = {
    "Bool": "PrependBool",
    "Uint8": "PrependUint8",
    "Uint16": "PrependUint16",
    "Uint32": "PrependUint32",
    "Uint64": "PrependUint64",
    "Int8": "PrependInt8",
    "Int16": "PrependInt16",
    "Int32": "PrependInt32",
    "Int64": "PrependInt64",
    "Float32": "PrependFloat32",
    "Float64": "PrependFloat64",
}


SCALAR_TYPES_TO_FLAGS: Final[dict[ScalarType, type]] = {
    "Float32": Float32Flags,
    "Float64": Float64Flags,
    "Uint8": Uint8Flags,
    "Uint16": Uint16Flags,
    "Uint32": Uint32Flags,
    "Uint64": Uint64Flags,
    "Int8": Int8Flags,
    "Int16": Int16Flags,
    "Int32": Int32Flags,
    "Int64": Int64Flags,
    "Bool": BoolFlags,
}



"""Types used internally."""
from enum import IntEnum

from attrs import NOTHING, AttrsInstance, fields, frozen, has

from ._consts import ENUM_TYPE_TO_SCALAR_TYPE, HELPER_TYPE_TO_SCALAR_TYPE, NoneType
from ._types import (
    FieldName,
    MaybeDefault,
    Optionality,
    PythonScalarType,
    ScalarType,
    SlotNumber,
    UnionMapping,
)
from .types import (
    Float,
    Float64,
    Int8,
    Int8Enum,
    Int16,
    Int16Enum,
    Int32,
    Int32Enum,
    Int64,
    Int64Enum,
    ScalarMarker,
    Uint8,
    Uint8Enum,
    Uint16,
    Uint16Enum,
    Uint32,
    Uint32Enum,
    Uint64,
    Uint64Enum,
    UnionVal,
)
from .typing import (
    get_annotation_and_base,
    get_optional_arg,
    get_union_args,
    is_generic_subclass,
    is_subclass,
    resolve_types,
)


@frozen
class FlatbufferTable:
    cl: type[AttrsInstance]
    num_slots: int
    inlines: list[tuple[FieldName, ScalarType, SlotNumber, MaybeDefault]]
    enums: list[tuple[FieldName, ScalarType, SlotNumber, MaybeDefault]]
    strings: list[tuple[FieldName, SlotNumber, Optionality]]
    byte_fields: list[tuple[FieldName, SlotNumber, Optionality]]
    tables: list[tuple[FieldName, type, SlotNumber, Optionality]]
    lists_of_strings: list[tuple[FieldName, SlotNumber, Optionality]]
    lists_of_scalars: list[
        tuple[FieldName, SlotNumber, PythonScalarType, ScalarType, Optionality]
    ]
    lists_of_enums: list[
        tuple[FieldName, SlotNumber, PythonScalarType, ScalarType, Optionality]
    ]
    lists_of_tables: list[tuple[FieldName, type, SlotNumber, Optionality]]
    unions: list[tuple[FieldName, tuple[type, ...], UnionMapping, SlotNumber]]


def analyze(
    cl: type[AttrsInstance],
    field_overrides: list[tuple[FieldName, ScalarType, MaybeDefault]] = [],
    scalar_list_overrides: list[tuple[FieldName, ScalarType]] = [],
    union_overrides: dict[FieldName, UnionMapping] = {},
) -> FlatbufferTable:
    resolve_types(cl)
    strings: list[tuple[FieldName, SlotNumber, Optionality]] = []
    byte_fields: list[tuple[FieldName, SlotNumber, Optionality]] = []
    tables: list[tuple[FieldName, type, SlotNumber, Optionality]] = []
    lists_of_tables: list[tuple[FieldName, type, SlotNumber, Optionality]] = []
    lists_of_strings: list[tuple[FieldName, SlotNumber, Optionality]] = []
    lists_of_scalars: list[
        tuple[FieldName, SlotNumber, PythonScalarType, ScalarType, Optionality]
    ] = []
    lists_of_enums: list[
        tuple[FieldName, SlotNumber, PythonScalarType, ScalarType, Optionality]
    ] = []
    enums: list[FieldName, str, SlotNumber, MaybeDefault] = []
    inlines: list[FieldName, ScalarType, SlotNumber, MaybeDefault] = []
    unions: list[tuple[FieldName, tuple[type, ...], UnionMapping, SlotNumber]] = []
    next_slot_idx = 0

    overridden = {f[0]: f for f in field_overrides}
    overridden_sl: dict[FieldName, ScalarType] = {
        f[0]: f[1] for f in scalar_list_overrides
    }

    for field in fields(cl):
        if field.name in overridden:
            f = overridden[field.name]
            if issubclass(field.type, IntEnum):
                enums.append((field.name, f[1], next_slot_idx, f[2]))
            else:
                inlines.append((field.name, f[1], next_slot_idx, f[2]))
            next_slot_idx += 1
            continue

        ftype = field.type
        if ftype is str:
            strings.append((field.name, next_slot_idx, False))
        elif ftype is bytes:
            byte_fields.append((field.name, next_slot_idx, False))
        elif has(ftype):
            if field.name in union_overrides:
                # A single-class union.
                unions.append(
                    (
                        field.name,
                        (ftype,),
                        union_overrides[field.name],
                        next_slot_idx,
                    )
                )
                next_slot_idx += 1
            else:
                tables.append((field.name, field.type, next_slot_idx, False))
        elif (anb := get_annotation_and_base(ftype, UnionVal)) is not None and has(
            anb[1]
        ):
            # Single-class union.
            unions.append(
                (
                    field.name,
                    (ftype,),
                    _make_union_mapping((ftype,)),
                    next_slot_idx,
                )
            )
            next_slot_idx += 1
        elif o := get_optional_arg(ftype):
            # This is an optional field.
            if o is str:
                strings.append((field.name, next_slot_idx, True))
            elif o is bytes:
                byte_fields.append((field.name, next_slot_idx, True))
            elif has(o):
                if field.name in union_overrides:
                    unions.append(
                        (
                            field.name,
                            (o, NoneType),
                            union_overrides[field.name],
                            next_slot_idx,
                        )
                    )
                    next_slot_idx += 1
                else:
                    tables.append((field.name, o, next_slot_idx, True))
            elif (anb := get_annotation_and_base(o, UnionVal)) is not None and has(
                anb[1]
            ):
                unions.append(
                    (
                        field.name,
                        (anb[1], NoneType),
                        _make_union_mapping((o, NoneType)),
                        next_slot_idx,
                    )
                )
                next_slot_idx += 1
            elif is_generic_subclass(o, list):
                arg = o.__args__[0]
                if arg is str:
                    lists_of_strings.append((field.name, next_slot_idx, True))
                elif has(arg):
                    lists_of_tables.append((field.name, arg, next_slot_idx, True))
                elif arg in (
                    bool,
                    Uint8,
                    Uint16,
                    Uint32,
                    Uint64,
                    Int8,
                    Int16,
                    Int32,
                    int,
                    Float,
                    Float64,
                ):
                    lists_of_scalars.append(
                        (
                            field.name,
                            next_slot_idx,
                            arg,
                            overridden_sl.get(
                                field.name, HELPER_TYPE_TO_SCALAR_TYPE[arg]
                            ),
                            True,
                        )
                    )
                elif issubclass(
                    arg,
                    (
                        Uint8Enum,
                        Uint16Enum,
                        Uint32Enum,
                        Uint64Enum,
                        Int8Enum,
                        Int16Enum,
                        Int32Enum,
                        Int64Enum,
                    ),
                ):
                    for enum_type, scalar_type in ENUM_TYPE_TO_SCALAR_TYPE.items():
                        if enum_type in arg.__mro__:
                            break
                    lists_of_enums.append(
                        (
                            field.name,
                            next_slot_idx,
                            arg,
                            overridden_sl.get(field.name, scalar_type),
                            True,
                        )
                    )
                else:
                    raise TypeError(f"Cannot handle {field.name} {ftype}")
            else:
                raise TypeError(f"Cannot handle {field.name} {ftype}")
        elif u := get_union_args(ftype):
            unions.append(
                (
                    field.name,
                    u,
                    union_overrides[field.name]
                    if field.name in union_overrides
                    else _make_union_mapping(u),
                    next_slot_idx,
                )
            )
            next_slot_idx += 1
        elif is_generic_subclass(ftype, list):
            arg = ftype.__args__[0]
            if arg is str:
                lists_of_strings.append((field.name, next_slot_idx, False))
            elif has(arg):
                lists_of_tables.append((field.name, arg, next_slot_idx, False))
            elif arg in HELPER_TYPE_TO_SCALAR_TYPE:
                lists_of_scalars.append(
                    (
                        field.name,
                        next_slot_idx,
                        arg,
                        overridden_sl.get(field.name, HELPER_TYPE_TO_SCALAR_TYPE[arg]),
                        False,
                    )
                )
            elif issubclass(
                arg,
                (
                    Uint8Enum,
                    Uint16Enum,
                    Uint32Enum,
                    Uint64Enum,
                    Int8Enum,
                    Int16Enum,
                    Int32Enum,
                    Int64Enum,
                ),
            ):
                for enum_type, scalar_type in ENUM_TYPE_TO_SCALAR_TYPE.items():
                    if enum_type in arg.__mro__:
                        break
                lists_of_enums.append(
                    (
                        field.name,
                        next_slot_idx,
                        arg,
                        overridden_sl.get(field.name, scalar_type),
                        False,
                    )
                )
            else:
                raise TypeError(f"Cannot handle {field.name} {ftype}")
        elif is_subclass(
            ftype,
            (
                Uint8Enum,
                Uint16Enum,
                Uint32Enum,
                Uint64Enum,
                Int8Enum,
                Int16Enum,
                Int32Enum,
                Int64Enum,
            ),
        ):
            # Enums before scalars, since IntEnum is a subclass of int.
            for helper_type, scalar_type in ENUM_TYPE_TO_SCALAR_TYPE.items():
                if helper_type in ftype.__mro__:
                    break
            enums.append(
                (
                    field.name,
                    scalar_type,
                    next_slot_idx,
                    field.default if field.default is not NOTHING else list(ftype)[0],
                )
            )
        elif get_annotation_and_base(ftype, ScalarMarker) is not None or ftype in (
            bool,
            int,
            float,
        ):
            inlines.append(
                (
                    field.name,
                    HELPER_TYPE_TO_SCALAR_TYPE[ftype],
                    next_slot_idx,
                    field.default,
                )
            )
        else:
            raise TypeError(f"Cannot handle {ftype}")

        next_slot_idx += 1
    return FlatbufferTable(
        cl,
        next_slot_idx,
        inlines,
        enums,
        strings,
        byte_fields,
        tables,
        lists_of_strings,
        lists_of_scalars,
        lists_of_enums,
        lists_of_tables,
        unions,
    )


def _make_union_mapping(union_types: tuple[type, ...]) -> UnionMapping:
    """Create a mapping of values to union members for use by the structuring logic."""
    res = {}

    curr_ix = 1
    for t in union_types:
        if t is NoneType:
            continue
        if union_val_and_base := get_annotation_and_base(t, UnionVal):
            union_val, base = union_val_and_base
            curr_ix = union_val
            t = base
        res[curr_ix] = t
        curr_ix += 1

    return res



from pathlib import Path

import click

from .renderer import render


@click.command()
@click.argument("input", type=click.Path(path_type=Path))
@click.argument("output", type=click.Path(path_type=Path))
@click.option("--gen-namespace-exports/--no-gen-namespace-exports", default=False)
def main(input: Path, output: Path, gen_namespace_exports: bool) -> None:
    render(input, output, gen_namespace_exports)


if __name__ == "__main__":
    main()
